var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
import clsx from 'clsx';
import { useId } from 'react';
import { InteractiveSection } from './InteractiveSection';
import { Markdown } from './Markdown';
import { checkIsReference, resolveDescription } from './utils';
import { stringifyOpenAPI } from './stringifyOpenAPI';
import { OpenAPISchemaName } from './OpenAPISchemaName';
import { OpenAPIDisclosure } from './OpenAPIDisclosure';
/**
 * Render a property of an OpenAPI schema.
 */
export function OpenAPISchemaProperty(props) {
    var _a;
    var schema = props.schema, _b = props.circularRefs, parentCircularRefs = _b === void 0 ? new Map() : _b, context = props.context, className = props.className;
    var id = useId();
    var parentCircularRef = parentCircularRefs.get(schema);
    var circularRefs = new Map(parentCircularRefs).set(schema, id);
    // Avoid recursing infinitely, and instead render a link to the parent schema
    var properties = parentCircularRef ? null : getSchemaProperties(schema);
    var alternatives = parentCircularRef
        ? null
        : getSchemaAlternatives(schema, new Set(circularRefs.keys()));
    if ((properties && !!properties.length) || schema.type === 'object') {
        return (<InteractiveSection id={id} className={clsx('openapi-schema', className)}>
                <OpenAPISchemaPresentation {...props}/>
                {properties && properties.length > 0 ? (<OpenAPIDisclosure context={context}>
                        <OpenAPISchemaProperties properties={properties} circularRefs={circularRefs} context={context}/>
                    </OpenAPIDisclosure>) : null}
            </InteractiveSection>);
    }
    if ((_a = alternatives === null || alternatives === void 0 ? void 0 : alternatives[0]) === null || _a === void 0 ? void 0 : _a.length) {
        return (<InteractiveSection id={id} className={clsx('openapi-schema', className)}>
                <OpenAPISchemaPresentation {...props}/>
                {alternatives[0].map(function (alternative, index) { return (<OpenAPISchemaAlternative key={"alternative-".concat(index)} schema={alternative} circularRefs={circularRefs} context={context}/>); })}
            </InteractiveSection>);
    }
    return (<InteractiveSection id={id} className={clsx('openapi-schema', className)}>
            <OpenAPISchemaPresentation {...props}/>
            {(properties && properties.length > 0) ||
            (schema.enum && schema.enum.length > 0) ||
            parentCircularRef ? (<>
                    {(properties === null || properties === void 0 ? void 0 : properties.length) ? (<OpenAPISchemaProperties properties={properties} circularRefs={circularRefs} context={context}/>) : null}
                    {parentCircularRef ? (<OpenAPISchemaCircularRef id={parentCircularRef} schema={schema}/>) : null}
                </>) : null}
        </InteractiveSection>);
}
/**
 * Render a set of properties of an OpenAPI schema.
 */
export function OpenAPISchemaProperties(props) {
    var id = props.id, properties = props.properties, circularRefs = props.circularRefs, context = props.context;
    if (!properties.length) {
        return null;
    }
    return (<div id={id} className="openapi-schema-properties">
            {properties.map(function (property) { return (<OpenAPISchemaProperty key={property.propertyName} circularRefs={circularRefs} {...property} context={context}/>); })}
        </div>);
}
/**
 * Render a root schema (such as the request body or response body).
 */
export function OpenAPIRootSchema(props) {
    var schema = props.schema, context = props.context;
    // Avoid recursing infinitely, and instead render a link to the parent schema
    var properties = getSchemaProperties(schema);
    if (properties && properties.length > 0) {
        return <OpenAPISchemaProperties properties={properties} context={context}/>;
    }
    return (<OpenAPISchemaProperty schema={schema} context={context} className="openapi-schema-root"/>);
}
/**
 * Render a tab for an alternative schema.
 * It renders directly the properties if relevant;
 * for primitives, it renders the schema itself.
 */
function OpenAPISchemaAlternative(props) {
    var schema = props.schema, circularRefs = props.circularRefs, context = props.context;
    var id = useId();
    var subProperties = getSchemaProperties(schema);
    return (<OpenAPIDisclosure context={context}>
            <OpenAPISchemaProperties id={id} properties={subProperties !== null && subProperties !== void 0 ? subProperties : [{ schema: schema }]} circularRefs={subProperties ? new Map(circularRefs).set(schema, id) : circularRefs} context={context}/>
        </OpenAPIDisclosure>);
}
/**
 * Render a circular reference to a schema.
 */
function OpenAPISchemaCircularRef(props) {
    var id = props.id, schema = props.schema;
    return (<div className="openapi-schema-circular">
            Circular reference to <a href={"#".concat(id)}>{getSchemaTitle(schema)}</a>{' '}
            <span className="openapi-schema-circular-glyph">↩</span>
        </div>);
}
/**
 * Render the enum value for a schema.
 */
export function OpenAPISchemaEnum(props) {
    var enumValues = props.enumValues;
    return (<div className="openapi-schema-enum">
            <span>
                Options:{' '}
                {enumValues.map(function (value, index) { return (<span key={index} className="openapi-schema-enum-value">
                        <code>{"".concat(value)}</code>
                        {index < enumValues.length - 1 ? ', ' : ''}
                    </span>); })}
            </span>
        </div>);
}
export function OpenAPISchemaPresentation(props) {
    var schema = props.schema, propertyName = props.propertyName, required = props.required;
    var shouldDisplayExample = function (schema) {
        return (typeof schema.example === 'string' ||
            typeof schema.example === 'number' ||
            typeof schema.example === 'boolean' ||
            (Array.isArray(schema.example) && schema.example.length > 0) ||
            (typeof schema.example === 'object' &&
                schema.example !== null &&
                Object.keys(schema.example).length > 0));
    };
    var description = resolveDescription(schema);
    return (<div className="openapi-schema-presentation">
            <OpenAPISchemaName type={getSchemaTitle(schema)} propertyName={propertyName} required={required} deprecated={schema.deprecated}/>
            {schema['x-deprecated-sunset'] ? (<div className="openapi-deprecated-sunset openapi-schema-description openapi-markdown">
                    Sunset date:{' '}
                    <span className="openapi-deprecated-sunset-date">
                        {schema['x-deprecated-sunset']}
                    </span>
                </div>) : null}
            {description ? (<Markdown source={description} className="openapi-schema-description"/>) : null}
            {shouldDisplayExample(schema) ? (<div className="openapi-schema-example">
                    Example:{' '}
                    <code>
                        {typeof schema.example === 'string'
                ? schema.example
                : stringifyOpenAPI(schema.example)}
                    </code>
                </div>) : null}
            {schema.pattern ? (<div className="openapi-schema-pattern">
                    Pattern: <code>{schema.pattern}</code>
                </div>) : null}
            {schema.enum && schema.enum.length > 0 ? (<OpenAPISchemaEnum enumValues={schema.enum}/>) : null}
        </div>);
}
/**
 * Get the sub-properties of a schema.
 */
function getSchemaProperties(schema) {
    if (schema.allOf) {
        return schema.allOf.reduce(function (acc, subSchema) {
            var _a;
            var properties = (_a = getSchemaProperties(subSchema)) !== null && _a !== void 0 ? _a : [
                {
                    schema: subSchema,
                },
            ];
            return __spreadArray(__spreadArray([], acc, true), properties, true);
        }, []);
    }
    // check array AND schema.items as this is sometimes null despite what the type indicates
    if (schema.type === 'array' && !!schema.items) {
        var items = schema.items;
        var itemProperties = getSchemaProperties(items);
        if (itemProperties) {
            return itemProperties;
        }
        return [
            {
                propertyName: 'items',
                schema: items,
            },
        ];
    }
    if (schema.type === 'object' || schema.properties) {
        var result_1 = [];
        if (schema.properties) {
            Object.entries(schema.properties).forEach(function (_a) {
                var propertyName = _a[0], propertySchema = _a[1];
                result_1.push({
                    propertyName: propertyName,
                    required: Array.isArray(schema.required)
                        ? schema.required.includes(propertyName)
                        : undefined,
                    schema: propertySchema,
                });
            });
        }
        if (schema.additionalProperties) {
            var additionalProperties = schema.additionalProperties;
            result_1.push({
                propertyName: 'Other properties',
                schema: additionalProperties === true ? {} : additionalProperties,
            });
        }
        return result_1;
    }
    return null;
}
/**
 * Get the alternatives to display for a schema.
 */
export function getSchemaAlternatives(schema, ancestors) {
    if (ancestors === void 0) { ancestors = new Set(); }
    var downAncestors = new Set(ancestors).add(schema);
    if (schema.anyOf) {
        return [flattenAlternatives('anyOf', schema.anyOf, downAncestors), schema.discriminator];
    }
    if (schema.oneOf) {
        return [flattenAlternatives('oneOf', schema.oneOf, downAncestors), schema.discriminator];
    }
    if (schema.allOf) {
        // allOf is managed in `getSchemaProperties`
        return null;
    }
    return null;
}
function flattenAlternatives(alternativeType, alternatives, ancestors) {
    return alternatives.reduce(function (acc, alternative) {
        var _a;
        if (!!alternative[alternativeType] && !ancestors.has(alternative)) {
            return __spreadArray(__spreadArray([], acc, true), (((_a = getSchemaAlternatives(alternative, ancestors)) === null || _a === void 0 ? void 0 : _a[0]) || []), true);
        }
        return __spreadArray(__spreadArray([], acc, true), [alternative], false);
    }, []);
}
export function getSchemaTitle(schema, 
/** If the title is inferred in a oneOf with discriminator, we can use it to optimize the title */
discriminator) {
    var _a;
    if (schema.title) {
        // If the schema has a title, use it
        return schema.title;
    }
    // Try using the discriminator
    if ((discriminator === null || discriminator === void 0 ? void 0 : discriminator.propertyName) && schema.properties) {
        var discriminatorProperty = schema.properties[discriminator.propertyName];
        if (discriminatorProperty && !checkIsReference(discriminatorProperty)) {
            if (discriminatorProperty.enum) {
                return discriminatorProperty.enum.map(function (value) { return value.toString(); }).join(' | ');
            }
        }
    }
    // Otherwise try to infer a nice title
    var type = 'any';
    if (schema.enum) {
        type = 'enum';
        // check array AND schema.items as this is sometimes null despite what the type indicates
    }
    else if (schema.type === 'array' && !!schema.items) {
        type = "".concat(getSchemaTitle(schema.items), "[]");
    }
    else if (Array.isArray(schema.type)) {
        type = schema.type.join(' | ');
    }
    else if (schema.type || schema.properties) {
        type = (_a = schema.type) !== null && _a !== void 0 ? _a : 'object';
        if (schema.format) {
            type += " ".concat(schema.format);
        }
    }
    else if ('anyOf' in schema) {
        type = 'any of';
    }
    else if ('oneOf' in schema) {
        type = 'one of';
    }
    else if ('allOf' in schema) {
        type = 'all of';
    }
    else if ('not' in schema) {
        type = 'not';
    }
    if (schema.nullable) {
        type = "nullable ".concat(type);
    }
    return type;
}
